"""This module defines analytic functions.

Sorting
    * ``one_dim_sort()``: One-dimensional sort.
    * ``two_dim_sort()``: Two-dimensional sort.

Time-Series Analysis
    * ``time_series_average()``: Calculate time-series mean and t-value.
    * ``rolling_beta()``: Run rolling OLS on a panel data.

Cross-sectional Analysis
    * ``crosssectional_regression()``: Run cross-sectional OLS and calculate the time-series means and t-values of the coefficients.

Portfolio Analysis
    * ``make_future_return()``: Compute future returns.
    * ``weighted_mean()``: Calculate weighted mean.
    * ``make_quantile_portfolios()``: Make quantile portfolios.
"""
import numpy as np
import pandas as pd
import statsmodels.api as sm

from pyanomaly.globals import *
from pyanomaly.numba_support import *
from pyanomaly.datatools import *
from pyanomaly.portfolio import Portfolio, Portfolios
from pyanomaly.multiprocess import multiprocess

################################################################################
#
# ONE-DIM / TWO-DIM SORT
#
################################################################################

def relabel_class(data, labels=None, axis=0, level=-1, col=None):
    """Relabel classes in columns/indexes or in a column of `data`.

    The existing labels should be continuous integers starting from 0.
    The `data` is relabeled in-place.

    Args:
        data: DataFrame to be relabeled.
        labels: New class labels.
        axis: 1: index, 2: column.
        level: Level of index/column to be relabeled.
        col: Column name. If column name is given, `axis` and `level` are ignored.
    """

    if labels is None:
        return

    name_mapping = {i: l for i, l in enumerate(labels)}

    if col is None:
        data.rename(name_mapping, axis=axis, level=level, inplace=True)
    else:
        for i, l in enumerate(labels):
            data.loc[data[col] == i, col] = l


def append_long_short(data, level=-1, l_label=None, s_label=None, ls_label=None):
    """Add long-short to `data`.

    Long-short is defined as (first group - last group) in each date.
    If labels are not given, long-short will be (class 0 - class N-1), where N is the number of classes,
    and the class label of the long-short is set to N.

    Args:
        data: DataFrame with index = date/class1/class2/....
        level: Index level to make long-short on. Default to the last level.
        l_label: Label of the long class. If None, `l_label = 0`.
        s_label: Label of the short class. If None, `s_label = num. classes - 1`.
        ls_label: Label of the long-short. If None, `ls_label = num. classes`.

    Returns:
        `data` with long-short appended.
    """

    data2 = data.swaplevel(-1, level).unstack()
    classes = list(data2.columns.get_level_values(-1).unique())
    l_label = l_label or 0
    s_label = s_label or len(classes) - 1
    ls_label = ls_label or len(classes)

    for col in data:
        if isinstance(col, tuple):
            data2[(*col, ls_label)] = data2[(*col, l_label)] - data2[(*col, s_label)]
        else:
            data2[(col, ls_label)] = data2[(col, l_label)] - data2[(col, s_label)]

    return data2.stack(dropna=True).swaplevel(-1, level)[data.columns]

# def append_long_short(data):
#
#     def append_long_short_(data):
#         date = data.index[0][0]
#         if len(data.index.levshape) == 3:
#             class1 = data.index[0][1]
#             ls = data.loc[(date, class1, l_label)] - data.loc[(date, class1, s_label)]
#             ls.name = (date, class1, ls_label)
#         else:
#             ls = data.loc[(date, l_label)] - data.loc[(date, s_label)]
#             ls.name = (date, ls_label)
#         return data.append(ls)
#
#     n_classes = len(data.index.get_level_values(-1).unique())
#     l_label = 0
#     s_label = n_classes - 1
#     ls_label = n_classes
#
#     return groupby_apply(data.groupby(level=0), append_long_short_)


def one_dim_sort(data, class_col, target_cols=None, weight_col=None, function='mean', add_long_short=True):
    """One-dimensional sort.

    This function assumes that `data` has already been sorted/grouped and the class labels are given
    in `class_col` column. Aggregate `data[target_cols]` using `data[class_col]` and return aggregated results.

    Args:
        data: DataFrame to be grouped. Index must be date/id.
        class_col: Class label column.
        target_cols: (List of) column(s) to aggregate. If None, `target_cols` = all numeric columns of `data`.
        weight_col: Weight column. If given, weighted mean is returned.
        function: Aggregate function, e.g., 'sum', 'mean', 'count', or a list of functions.
            If `function` != 'mean', `weight_col` is ignored.
        add_long_short (bool): Add long-short to the output.

    Returns:
        Aggregated data with index = date/class, columns = `target_cols`.
    """

    date_col = data.index.names[0]
    if target_cols is None:  # target_cols = all columns in data except for non-numeric ones.
        target_cols = list(data.columns[(data.dtypes == float) | (data.dtypes == int)])
    else:
        target_cols = to_list(target_cols)

    if (function == 'mean') and weight_col:  # value_weight
        ret_data = weighted_mean(data, target_cols, weight_col, [date_col, class_col])
    elif isinstance(function, str):  # single function
        gb = data.groupby([date_col, class_col])[target_cols]
        ret_data = eval(f'gb.{function}()')
    else:  # list of functions
        gb = data.groupby([date_col, class_col])[target_cols]
        ret_data = eval(f'gb.agg({function})')

    # Remove unclassified data.
    # ret_data = ret_data[~ret_data.index.get_level_values(class_col).isin([np.nan, ''])]

    if add_long_short:
        ret_data = append_long_short(ret_data)

    return ret_data


def two_dim_sort(data, class_col1, class_col2, target_cols=None, weight_col=None, function='mean',
                 add_long_short=True, output_dim=1):
    """Two-dimensional sort.

    This function assumes that `data` has already been sorted/grouped and the class labels are given
    in `class_col1` and `class_col2` columns. Aggregate `data[target_cols]` using `data[class_col1, class_col2]` and
    return aggregated results.

    Args:
        data: Data to be grouped. Index must be date/id.
        class_col1/2: Class label column for the 1st (2nd) dimension.
        target_cols: (List of) column(s) to aggregate. If None, `target_cols` = all numeric columns of `data`.
        weight_col: Weight column. If given, weighted mean is returned.
        function: Aggregate function, e.g., 'sum', 'mean', 'count'. If it is other than 'mean', `weight_col` is ignored.
        add_long_short (bool): Add long-short to the output.
        output_dim: If 1, output is a DataFrame with index = date/class1/class2; if 2, output is a DataFrame with
            index = date/class1 and column = class2.

    Returns:
        Aggregated data (DataFrame or dict of DataFrames).

        * If `output_dim` = 1, index = date/class1/class2 and columns = `target_cols`.
        * If `output_dim` = 2 and `len(target_cols)` = 1, index = date/class1 and columns = class2.
        * If `output_dim` = 2 and `len(target_cols)` > 1, output is dict with keys = `target_cols` and
          values = DataFrames (as in the second case).
    """

    date_col = data.index.names[0]
    if target_cols is None:  # target_cols = all columns in data except for non-numeric ones.
        target_cols = list(data.columns[(data.dtypes == float) | (data.dtypes == int)])
    else:
        target_cols = to_list(target_cols)

    if (function == 'mean') and weight_col:  # value_weight
        ret_data = weighted_mean(data, target_cols, weight_col, [date_col, class_col1, class_col2])
    else:  # equal-weight
        # gb = data[[class_col1, class_col2] + target_cols].groupby([date_col, class_col1, class_col2])
        gb = data.groupby([date_col, class_col1, class_col2])[target_cols]
        ret_data = eval(f'gb.{function}()')

    # Remove unclassified data.
    # ret_data = ret_data[~ret_data.index.get_level_values(class_col1).isin([np.nan, ''])]
    # ret_data = ret_data[~ret_data.index.get_level_values(class_col2).isin([np.nan, ''])]

    if add_long_short:
        ret_data = append_long_short(ret_data, level=-2)
        ret_data = append_long_short(ret_data, level=-1)

    if output_dim == 1:  # 1D with index=date/class1/class2
        return ret_data  # .to_frame(name=target_col)
    else:  # 2D with index=date/class1, column=class2
        index = ret_data.index.get_level_values(-2).unique()
        columns = ret_data.index.get_level_values(-1).unique()
        if len(target_cols) == 1:
            return ret_data[target_cols[0]].unstack().reindex(index, level=-1)[
                columns]  # this is necessary as unstack() sort index and columns.
        else:
            ret_dict = {}
            for target_col in target_cols:
                # Below, reindex is necessary as unstack() sort index and columns.
                ret_dict[target_col] = ret_data[target_col].unstack().reindex(index, level=-1)[columns]
            return ret_dict


################################################################################
#
# TIME-SERIES ANALYSIS
#
################################################################################

def t_value(data, cov_type='nonrobust', cov_kwds=None):
    """Calculate t-value for each column under H0: x = 0.

    Args:
        data: DataFrame with each column containing samples.
        cov_type: See `sm.OLS.fit()`, eg, 'HAC' for Newey-West.
        cov_kwds: See `sm.OLS.fit()`, eg, {'maxlags: 12} for cov_type = 'HAC'.

    Returns:
        Series of t-values with index = `data.columns`.
    """

    x = np.ones([len(data), 1])
    if isinstance(data, pd.Series):
        est = sm.OLS(data, x, missing='drop').fit(cov_type=cov_type, cov_kwds=cov_kwds)
        return est.tvalues[0]
    else:  # DataFrame
        res = {}
        for col in data.columns:
            est = sm.OLS(data[col], x, missing='drop').fit(cov_type=cov_type, cov_kwds=cov_kwds)
            res[col] = est.tvalues[0]
        res = pd.Series(res)
        res.index = res.index.set_names(data.columns.names)
        return res


def time_series_average(data, cov_type='nonrobust', cov_kwds=None):
    """Calculate time-series mean and t-value of each column (and for each group).

    `data` can be either a time-series data (index = date) or a panel data (index = date/group).
    If it is a panel, time-series mean and t-value are calculated for each group.

    Args:
        data (DataFrame): Data to analyze. If MultiIndex, the first index must be date.
        cov_type: See `sm.OLS.fit()`, eg, 'HAC' for Newey-West.
        cov_kwds: See `sm.OLS.fit()`, eg, {'maxlags: 12} for cov_type = 'HAC'.

    Returns:
        mean (DataFrame), t-value (DataFrame).

        * If data has MultiIndex, mean (t-value) has index = `data.index[1:]` and columns = `data.columns`.
        * Otherwise, mean (t-value) has index = `data.columns` and columns = 'mean' ('t-stat').
    """
    if not isinstance(data.index, pd.MultiIndex):  # single (date) index
        mean = data.mean().to_frame(name='mean')
        tval = t_value(data, cov_type, cov_kwds).to_frame(name='t-stat')
        return mean, tval
    else:  # MultiIndex.
        data = data.reset_index(level=0, drop=True)
        index = data.index.unique()

        mean = data.groupby(index.names).mean()
        tval = data.groupby(index.names).apply(t_value, cov_type, cov_kwds)
        if isinstance(data, pd.DataFrame):
            tval.columns.names = mean.columns.names
        return mean.reindex(index), tval.reindex(index)  # reindex to keep the order of index


def rolling_beta(data, window, minobs=None, endo_col=None, exog_cols=None):
    """Run rolling OLS on a panel, `data`.

    The `data` must have index = date/id and Rolling OLS is applied to each id.
    This is faster than `statsmodels.RollingOLS`, but the output is limited to
    coefficients, R2, and idiosyncratic volatility.

    Args:
        data: DataFrame with index = date/id, sorted on id/date.
        window: Window size.
        minobs: Minimum number of observations in the window. If observations < `minobs`, result is nan.
        endo_col: y column.
        exog_cols: X columns (without constant). If `endo_col = exog_cols = None`, `endo_col = data[:, 0]`,
            `exog_cols = data[:, 1:]`.

    Returns:
        Coefficients, R2, idiosyncratic volatility. These are NxK, Nx1, and Nx1 ndarrays, respectively,
        where N = `len(data)` and K = `len(exog_cols)+1`.
    """

    minobs = minobs or window
    data = data if endo_col is None else data[[endo_col] + exog_cols]

    @njit(error_model='numpy')
    def fcn(data):
        beta = np.full(data.shape, np.nan)
        r2 = np.full((data.shape[0], 1), np.nan)
        idio = np.full((data.shape[0], 1), np.nan)

        isnan = isnan1(data)
        for i in range(minobs - 1, data.shape[0]):
            i0 = i - min(i, window-1)
            sample = data[i0:i + 1]
            sample = sample[~isnan[i0:i + 1]]

            if sample.shape[0] < minobs:
                continue

            endo = sample[:, 0].copy()
            exog = add_constant(sample[:, 1:])

            try:
                beta[i, :] = (np.linalg.solve(exog.T @ exog, exog.T @ endo)).reshape(data.shape[1])
                pred = (beta[i, :] * exog).sum(1)
                r2[i] = 1 - ((endo - pred) ** 2).sum() / ((endo - endo.mean()) ** 2).sum()
                idio[i] = std(endo - pred)
            except:
                continue

        return beta, r2, idio

    return rolling_apply_np(data, fcn=fcn)



################################################################################
#
# CROSS-SECTIONAL ANALYSIS
#
################################################################################

def crosssectional_regression(data, endo_col, exog_cols, add_constant=True, cov_type='nonrobust', cov_kwds=None):
    """Run cross-sectional OLS on each date and calculate the time-series means and t-values of the coefficients.

    Args:
        data: DataFrame with index = date/id.
        endo_col: y column.
        exog_cols: List of X columns.
        add_constant (bool): Add constant to x.
        cov_type: See `sm.OLS.fit()`, eg, 'HAC' for Newey-West.
        cov_kwds: See `sm.OLS.fit()`, eg, {'maxlags: 12} for cov_type = 'HAC'.

    Returns:
        mean, tval, coefs.

        * mean (DataFrame): Time-series means of coefficients with index = ('const') + `exog_cols` and columns = 'mean'.
        * tval (DataFrame): t-values of coefficients with index = ('const') + `exog_cols` and columns = 't-stat'.
        * coefs (DataFrame): Coefficient time-series with index = dates and columns = ('const') + `exog_cols`.
    """
    data = data.dropna(subset=[endo_col] + exog_cols)

    # cross-sectional regression
    coefs = {}
    for k, v in data.groupby(level=0):
        y = v[endo_col]
        x = sm.add_constant(v[exog_cols]) if add_constant else v[exog_cols]
        if x.shape[0] <= x.shape[1]:  # If not enough sample for regression, skip.
            continue
        model = sm.OLS(y, x).fit()

        coefs[k] = model.params
    coefs = pd.DataFrame(coefs).T

    # time series means and t-values of the cross-sectional regression coefficients
    mean, tval = time_series_average(coefs, cov_type, cov_kwds)

    return mean, tval, coefs


################################################################################
#
# Portfolio Analysis
#
################################################################################

def make_future_return(ret, period=1):
    """Compute future returns.

    This is a simple function to compute `period`-period ahead return.

    See Also:
        ``Panel.cumret()``

    Args:
        ret: Series of returns with index = date or date/id.
        period: target period. Eg: If `period` = 3, 3-period ahead (cumulative) returns are returned. If `ret` contains monthly (daily)
            returns, the output will be 3-month (day) ahead returns.

    Returns:
        Series of future returns.
    """

    if period == 1:
        futret = ret.shift(-1)
    else:
        futret = np.exp(np.log(ret + 1).shift(-period).rolling(period).sum()) - 1

    if isinstance(ret.index, pd.MultiIndex):
        ids = ret.index.get_level_values(-1).values
        futret[ids != np.roll(ids, -period)] = np.nan

    return futret


def weighted_mean(data, target_cols, weight_col, group_cols):
    """Calculate weighted means of `data[target_cols]` for each group defined by `group_cols`.

    Args:
        data: DataFrame.
        target_cols: (List of) column(s) to calculate weighted-mean.
        weight_col: Weight column name, or Series or ndarray of weights.
        group_cols: (List of) grouping column(s).

    Returns:
        DataFrame of weighted mean with index = `group_cols` and columns = `target_cols`.

    If `data` is a panel with index = 'date'/'permno' and column 'ret' contains returns
    and 'me' contains market equity at t-1, weighted mean returns can be obtained as follows:

    >>> wmean = weighted_mean(data, 'ret', 'me', 'permno')
    """

    target_cols = to_list(target_cols)
    group_cols = to_list(group_cols)
    if is_iterable(weight_col):
        weight = weight_col
    else:
        weight = data[weight_col]

    numer = data[target_cols].mul(weight, axis=0)
    denom = pd.notnull(data[target_cols]).mul(weight, axis=0)  # set weight=0 where target data is null.
    da = data[data.columns.intersection(group_cols)]  # data necessary for groupby (to save memory)
    gb1 = pd.concat([da, numer], axis=1).groupby(group_cols)
    gb2 = pd.concat([da, denom], axis=1).groupby(group_cols)

    return gb1[target_cols].sum().divide(gb2[target_cols].sum())


# def wmean(x, w):
#     if x.ndim == 1:
#         return x.dot(w) / w[~np.isnan(x)].sum()
#     elif x.ndim == 2:
#         retval = np.full(x.shape[1], np.nan)
#         for i in range(x.shape[1]):
#             x_ = x[:, i]#.copy()
#             retval[i] = x_.dot(w) / w[~np.isnan(x_)].sum()
#
#         return retval
#
#
# def weighted_mean(data, target_cols, weight_col, group_cols):
#     """Calculate weighted means of `data[target_cols]` for each group defined by `group_cols`.
#
#     Args:
#         data: DataFrame.
#         target_cols: (List of) column(s) to calculate weighted-mean.
#         weight_col: Weight column name, or Series or ndarray of weights.
#         group_cols: (List of) grouping column(s).
#
#     Returns:
#         DataFrame of weighted mean with index = `group_cols`, columns = `target_cols`.
#
#     """
#
#     if is_iterable(weight_col):
#         data['__weight__'] = weight_col
#         weight_col = '__weight__'
#
#     index = []
#     wmeans = []
#     columns = to_list(group_cols) + to_list(target_cols) + [weight_col]
#     columns = list(data.columns.intersection(columns))
#     for k, g in data[columns].groupby(group_cols):
#         index.append(k)
#         wmeans.append(wmean(g[target_cols].values, g[weight_col].values))
#
#     values = np.concatenate(wmeans) if isinstance(wmeans[0], np.ndarray) else wmeans
#     index = pd.MultiIndex.from_tuples(index, names=group_cols) if is_iterable(group_cols) and len(group_cols) > 1 else pd.Index(index, name=group_cols)
#     retval = pd.DataFrame(values, columns=to_list(target_cols), index=index)
#     if '__weight__' in data:
#         del data['__weight__']
#
#     return retval


def make_position(data, ret_col, weight_col=None, pf_col=None, other_cols=None):
    """Make portfolio position data from a panel data.

    To construct and evaluate a portfolio using ``Portfolio`` class, position data is required.
    This function makes the position data from the input data, which is a panel of securities.
    The position data is generated via the following operations:

        - Shift dates forward so that the return at t is the return between t-1 and t
          and other variables are as of t-1.
        - Change column names as assumed in ``Portfolio``:

            - 'date': Date column.
            - 'id': Security id column.
            - 'ret': Return column.
            - 'wgt': Weight column.
        - Generate portfolio weights by normalizing the values given in `weight_col`.

    Args:
        data: DataFrame with index = date/id.
        ret_col: Return column of `data`.
        weight_col: Weight column of `data`. If None, constituents are equally weighted.
        pf_col: Portfolio column of `data`, i.e., a column with portfolio id (name) a security belongs to.
            This can be None if the input data is for one portfolio.
        other_cols: Other columns of `data` the user wants to keep in the position data.

    Returns:
        Position DataFrame with index = 'date' and columns = ['id', 'ret', 'wgt'] + `other_cols`.
    """

    columns = unique_list([ret_col, weight_col, pf_col, other_cols])
    position = data[columns].groupby(level=-1).shift(1)  # Shift data forward.
    position = position.dropna(subset=[ret_col])  # Drop nan returns.

    # Rename: date: 'date', id: 'id', ret_col: 'ret', weight_col: 'wgt'
    position.index = position.index.set_names(['date', 'id'])
    position.rename(columns={ret_col: 'ret'}, inplace=True)

    if weight_col is None:  # Equal-weight portfolio.
        position['wgt'] = 1.
    else:
        if other_cols and (weight_col in other_cols):  # keep weight_col.
            position['wgt'] = position[weight_col]
        else:  # rename weight_col.
            position.rename(columns={weight_col: 'wgt'}, inplace=True)

    if pf_col:  # Normalize weights per date/portfolio.
        position['wgt'] = position.wgt / position.groupby(['date', pf_col]).wgt.transform('sum')
    else:  # Normalize weights per date.
        position['wgt'] = position.wgt / position.groupby('date').wgt.transform('sum')

    return position.reset_index('id').sort_index()


def make_portfolio(data, ret_col, weight_col=None, costfcn=None, keep_position=True, name=''):
    """Construct a portfolio from panel data.

    This function creates portfolio position data from `data` and construct a portfolio from it.

    Args:
        data: DataFrame with index = date/id.
        ret_col: Return column of `data`.
        weight_col: Weight column of `data`. If None, constituents are equally weighted.
        costfcn: Transaction cost. See ``pyanomaly.Portfolio`` for details.
        keep_position: If True, keep the position information. If position information is not needed,
            set this to False to save memory.
        name: Portfolio name.

    Returns:
        ``Portfolio`` object.
    """

    position = make_position(data, ret_col, weight_col)
    return Portfolio(name, position, costfcn=costfcn, keep_position=keep_position)


def make_long_short_portfolio(lposition, sposition, ls_wgt=(1, -1), rf=None, costfcn=None, keep_position=True, name='H-L'):
    """Make a long-short portfolio.

    Args:
        lposition (DataFrame): Long position data.
        sposition (DataFrame): Short position data.
        ls_wgt: Long-short weights. (1, -1) means 1:1 long-short. If None, long-short portfolio is not constructed.
        rf: Series or DataFrame of risk-free rates. The index should be date.
        costfcn: Transaction cost. See ``pyanomaly.Portfolio`` for details.
        keep_position: If True, keep the position information. If position information is not needed,
            set this to False to save memory.
        name: Portfolio name.

    Returns:
        ``Portfolio`` object.
    """
    # Copy data as their values ('wgt') are changed.
    lposition = lposition.copy()
    sposition = sposition.copy()

    lposition['wgt'] *= ls_wgt[0]
    sposition['wgt'] *= ls_wgt[1]
    position = pd.concat([lposition, sposition]).sort_index()
    portfolio = Portfolio(name, position, rf=rf, costfcn=costfcn, keep_position=keep_position)
    return portfolio


def make_quantile_portfolios(position, pf_col, ls_wgt=(1, -1), rf=None, costfcn=None, keep_position=True, labels=None):
    """Make quantile portfolios.

    This function makes quantile portfolios and the long-short portfolio from position data.

    Args:
        position (DataFrame): Position data. See ``make_position()`` for the format.
        pf_col: Column of `position` that maps a security with quantiles (portfolios).
        ls_wgt: Long-short weights. (1, -1) means 1:1 long-short. If None, long-short portfolio is not constructed.
        rf: Series or DataFrame of risk-free rates. The index should be date.
        costfcn: Transaction cost. See ``pyanomaly.Portfolio`` for details.
        keep_position: If True, keep the position information. If position information is not needed,
            set this to False to save memory.
        labels: Portfolio names. If None, the values in `pf_col` are used.

    Returns:
        ``Portfolios`` object.
    """

    log(f'Making quantile portfolios for {pf_col}...')

    n_class = len(position[pf_col].unique())
    if position[pf_col].isna().any():
        n_class -= 1  # Exclude unclassified (nan) data.

    portfolios = Portfolios()
    for cls in range(n_class):
        pf_name = str(cls) if labels is None else labels[cls]
        cposition = position.loc[position[pf_col] == cls, position.columns.difference([pf_col])]
        portfolio = Portfolio(pf_name, cposition, rf=rf, costfcn=costfcn, keep_position=keep_position)
        portfolios.add(portfolio)

    # Long-short
    if ls_wgt is not None:
        pf_name = str(n_class) if labels is None else labels[-1]
        lposition = position[position[pf_col] == 0]
        sposition = position[position[pf_col] == n_class - 1]
        portfolio = make_long_short_portfolio(lposition, sposition, ls_wgt, rf, costfcn, keep_position, pf_name)
        portfolios.add(portfolio)

    log('Quantile portfolios created...')
    return portfolios


if __name__ == '__main__':
    os.chdir('../')






class struct(dict):
    """dict that allows '.' expression.

    Examples:
        >>> x = struct(
        ...    a='py',
        ...    b='anomaly'
        ...    )
        >>> x['a']
        py
        >>> x.a
        py
    """

    def __init__(self, *args, **kwargs):
        if len(args) and type(args[0]) == dict:
            self.update(args[0])
        else:
            super().__init__(*args, **kwargs)
        for k, v in self.items():
            self.__dict__[k] = v

    def __setitem__(self, k, value):
        self.__dict__[k] = value
        super().__setitem__(k, value)

    def __setattr__(self, k, value):
        self.__dict__[k] = value
        if k in self.keys():  # update dict only when the key already exists.
            super().__setitem__(k, value)

    def __copy__(self):
        cls = self.__class__
        cpy = cls.__new__(cls)
        cpy.update(self)
        cpy.__dict__.update(self.__dict__)
        return cpy

    def copy(self):
        return self.__copy__()


class iterstruct(struct):
    """Iterable struct."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._idx = 0

    def __getitem__(self, k):
        if type(k) == str:
            return super().__getitem__(k)
        else:  # int
            return list(self.values())[k]

    def __iter__(self):
        return self

    def __next__(self):
        if self._idx == len(self):
            self._idx = 0
            raise StopIteration
        else:
            self._idx += 1
            return list(self.values())[self._idx - 1]




import os
import numpy as np
import pandas as pd
from .datatypes import struct

########################################
# File IO
########################################
config = struct(
    input_dir='./input/',
    output_dir='./output/',
    mapping_file_path='./mapping.xlsx',

    factors_monthly_fname='factors_monthly',
    factors_daily_fname='factors_daily',
    # factor_name in pyanomaly: factor_name in monthly(daily)_factors file.
    factor_names={
        'rf': 'rf',
        'mktrf': 'mktrf',
        'smb_ff': 'smb_ff',
        'hml': 'hml',
        'inv': 'inv',
        'roe': 'roe',
        'smb_hxz': 'smb_hxz',
    },
    # True to replicate JKP version (not completely). Verification purpose only. Will be deprecated.
    REPLICATE_JKP=False

)

########################################
# Display/Stdout
########################################
# pandas
pd.set_option('display.min_rows', 100, 'display.max_rows', 100, 'display.max_columns', 1000,
              'display.float_format', '{:.3f}'.format, 'display.width', 320, 'display.max_colwidth', 100)

# numpy
np.seterr(divide='ignore', invalid='ignore')  # Silence divide by 0) or nan warning.

# matplotlib
# try:  # try to use latex font when possible.
#     plt.rc('text', usetex=True)
#     plt.rc('font', family='serif', size=10)
# except:
#     pass


########################################
# Other configurations
########################################



from .datatypes import iterstruct
from .config import *
from .log import *
from .util import *

# weight_scheme
VW = 'vw'  # value-weight
EW = 'ew'  # equal-weight

# data frequency
ANNUAL = 1
QUARTERLY = 4
MONTHLY = 12
DAILY = 365

# pyanomaly frequency to pandas frequency symbols.
freq_map = {
    ANNUAL: 'Y',
    QUARTERLY: 'Q',
    MONTHLY: 'M',
    DAILY: 'D'
}

# Colour codes
# https://github.com/vega/vega/wiki/Scales#scale-range-literals
COLOR10 = iterstruct(
    blue = '#1f77b4',
    orange = '#ff7f0e',
    green = '#2ca02c',
    red = '#d62728',
    purple = '#9467bd',
    brown = '#8c564b',
    pink = '#e377c2',
    grey = '#7f7f7f',
    mustard = '#bcbd22',
    sky = '#17becf'
)

COLOR20 = iterstruct(
    blue1 = '#1f77b4',
    blue0 = '#aec7e8',
    orange1 = '#ff7f0e',
    orange0 = '#ffbb78',
    green1 = '#2ca02c',
    green0 = '#98df8a',
    red1 = '#d62728',
    red0 = '#ff9896',
    purple1 = '#9467bd',
    purple0 = '#c5b0d5',
    brown1 = '#8c564b',
    brown0 = '#c49c94',
    pink1 = '#e377c2',
    pink0 = '#f7b6d2',
    grey1 = '#7f7f7f',
    grey0 = '#c7c7c7',
    mustard1 = '#bcbd22',
    mustard0 = '#dbdb8d',
    sky1 = '#17becf',
    sky0 = '#9edae5',
)




import numpy as np
import pandas as pd

from .globals import *


# def table_regression(mean, tval, axis=0):
#     if axis == 0: # Series. mean[0], tval[0], mean[1], tval[1], ...
#         tval = tval.rename({i: '  t-stat' for i in tval.index})
#         tab = pd.concat([mean.iloc[:, 0], tval.iloc[:, 0]])
#         idx = []
#         for i in range(len(mean)):
#             idx.append(i)
#             idx.append(len(mean) + i)
#
#         return tab.iloc[idx]
#     else:  # DataFrame. columns = ['mean', 't-stat']
#         return pd.concat([mean, tval], axis=1)




"""This module defines `Panel` class, which serves as the base class for panel data analysis."""

import numpy as np
import pandas as pd
import json
import statsmodels.api as sm
from statsmodels.regression.rolling import RollingOLS

from pyanomaly.globals import *
from pyanomaly.datatools import *
from pyanomaly.analytics import rolling_beta
from pyanomaly.fileio import write_to_file, read_from_file


class Panel:
    """Base class for panel data.

    Data is stored in the attribute `data`. `data` should have index = date/id, i.e., the first index should be
    time-series identifier and the second index should be cross-section identifier, and always be sorted on id/date.
    The date index must be of datetime type.

    Args:
        alias: Characteristic column name in ``mapping.xlsx``. If None, function names (without 'c\_') are used as the
            characteristic names.
        data: DataFrame with index = date/id, sorted on id/date.
        freq: Frequency of `data`. ANNUAL, QUARTERLY, MONTHLY, or DAILY.
        base_freq: Frequency of `data` values. funda: ANNUAL, fundq: QUARTERLY, crspm: MONTHLY, crspd: DAILY.
            For example, if funda is populated monthly, `freq` = MONTHLY and `base_freq` = ANNUAL.
            If None, `base_freq` = `freq`.

    Attributes:
        data: DataFrame with index = date/id, sorted on id/date that stores a panel data.
        freq: Frequency of `data`. ANNUAL, QUARTERLY, MONTHLY, or DAILY.
        base_freq: Frequency of `data` values. ANNUAL, QUARTERLY, MONTHLY, or DAILY.
        char_map: Dictionary to map functions with aliases. `char_map['alias']` returns its corresponding function name.
        reverse_map: Function to alias mapping. `reverse_map['function']` returns its corresponding alias.
    """

    def __init__(self, alias=None, data=None, freq=MONTHLY, base_freq=None):
        self.data = data
        self.freq = freq
        self.base_freq = base_freq or freq

        # alias -> function
        alias_ = alias or 'function'
        char_map = pd.read_excel(config.mapping_file_path)
        char_map.dropna(subset=[alias_], inplace=True)
        self.char_map = char_map.set_index(alias_, drop=False)['function'].to_dict()

        if alias is None:  # Add new characteristic functions not in the mapping file yet.
            char_fncs = [fcn[2:] for fcn in dir(self) if callable(getattr(self, fcn)) and fcn[:2] == 'c_']
            for fcn in char_fncs:
                self.char_map[fcn] = fcn

        # function -> alias
        self.reverse_map = {v: k for k, v in self.char_map.items()}

    def rename(self, to_alias=True):
        """Rename `data` columns.

        Args:
            to_alias: If True, rename firm characteristic columns from function names to aliases, and vice-versa.
        """

        if to_alias:  # function -> alias
            map = {col: self.reverse_map[col] for col in self.data if col in self.reverse_map}
        else:  # alias -> function
            map = {col: self.char_map[col] for col in self.data if col in self.char_map}

        self.data.rename(columns=map, inplace=True)

    def date_col(self):
        """Get date index name."""

        return self.data.index.names[0]

    def id_col(self):
        """Get ID index name."""

        return self.data.index.names[-1]

    def date_values(self):
        """Get date index values."""

        return self.data.index.get_level_values(0)

    def id_values(self):
        """Get ID index values."""

        return self.data.index.get_level_values(-1)

    def inspect_data(self, option=['summary']):
        """Inspect data.

        See ``datatools.inspect_data()`` for details.
        """

        inspect_data(self.data, option)

    def save(self, fname=None, fdir=None, other_columns=None):
        """Save this object.

        Parameters are saved to a json file and `data` is saved to a pickle file.

        Args:
            fname: File name without extension. If None, `fname` = class name.
            fdir: Directory to save the file. If None, `fdir` = `config.output_dir`.
            other_columns: List of columns to save other than firm characteristics. If given, characteristic columns
                plus `other_columns` are saved. If None, all columns are saved. Note that unsaved columns will be deleted
                from `self.data`.
        """

        fname = fname or type(self).__name__.lower()  # lower-case class name
        fdir = fdir or config.output_dir
        fpath = fdir + ('' if fdir[-1] == '/' else '/') + fname + '.json'

        # Save data.
        if other_columns is not None:
            columns = other_columns + list(self.reverse_map.keys())
            keep_columns(self.data, columns)

        self.rename()  # To aliases.
        write_to_file(self.data, fname, fdir)
        self.rename(False)  # Back to function names.

        # Save parameters.
        params = {
            'freq': self.freq,
            'base_freq': self.base_freq,
            'char_map': self.char_map,
            'reverse_map': self.reverse_map,
        }
        with open(fpath, 'w') as fh:
            json.dump(params, fh, indent=4)

    def load(self, fname=None, fdir=None):
        """Load this object from a file.

        Args:
            fname: File name without extension. If None, `fname` = class name.
            fdir: Directory to read the file from. If None, `fdir` = `config.output_dir`.
        """

        fname = fname or type(self).__name__.lower()  # lower-case class name
        fdir = fdir or config.output_dir
        fpath = fdir + ('' if fdir[-1] == '/' else '/') + fname + '.json'

        # Load parameters.
        with open(fpath, 'r') as fh:
            params = json.load(fh)

        self.freq = params['freq']
        self.base_freq = params['base_freq']
        self.char_map = params['char_map']
        self.reverse_map = params['reverse_map']

        # Load data.
        self.data = read_from_file(fname, fdir)
        self.rename(False)

    def _filter_data(self, filter, keep_row=False):
        column = filter[0]
        condition = filter[1]
        value = filter[2]

        if condition == '==':
            if keep_row:
                self.data.loc[self.data[column] != value, :] = None
            else:
                self.data = self.data[self.data[column] == value]
        elif condition == '!=':
            if keep_row:
                self.data.loc[self.data[column] == value, :] = None
            else:
                self.data = self.data[self.data[column] != value]
        elif condition == '>':
            if keep_row:
                self.data.loc[self.data[column] <= value, :] = None
            else:
                self.data = self.data[self.data[column] > value]
        elif condition == '<':
            if keep_row:
                self.data.loc[self.data[column] >= value, :] = None
            else:
                self.data = self.data[self.data[column] < value]
        elif condition == '>=':
            if keep_row:
                self.data.loc[self.data[column] < value, :] = None
            else:
                self.data = self.data[self.data[column] >= value]
        elif condition == '<=':
            if keep_row:
                self.data.loc[self.data[column] > value, :] = None
            else:
                self.data = self.data[self.data[column] <= value]
        elif condition == 'in':
            if keep_row:
                self.data.loc[~self.data[column].isin(value), :] = None
            else:
                self.data = self.data[self.data[column].isin(value)]
        elif condition == 'not in':
            if keep_row:
                self.data.loc[self.data[column].isin(value), :] = None
            else:
                self.data = self.data[~self.data[column].isin(value)]
        else:
            raise ValueError(f'Unrecognized condition: {condition}')

    def filter(self, filters, keep_row=False):
        """Filter data using the conditions defined by `filters`.

        A filter is a tuple of three elements:

            * filter[0]: column to apply the filter to.
            * filter[1]: filter condition: '==', '!=', '>', '<', '>=', '<=', 'in', or 'not in'.
            * filter[2]: rhs value.

        If `filters` is a list of filters, they are applied sequentially.

        For example, to remove `data['x'] < 10`,

            >>> panel.filter_data(('x', '>=', 10))

        This is equivalent to

            >>> panel.data = panel.data[panel.data['x'] >= 10]

        Args:
            filters: A filter or list of filters.
            keep_row: Whether to keep or remove the filtered out rows. If True, the values of the filtered rows are
                set to None.
        """

        if isinstance(filters[0], str):
            filters = [filters]

        for filter in filters:
            self._filter_data(filter, keep_row)

    def show_available_functions(self):
        """Display all functions for firm characteristics."""

        fcns = {fcn[2:]: getattr(self, fcn).__doc__ for fcn in dir(self) if callable(getattr(self, fcn)) and fcn[:2] == 'c_'}
        s = pd.Series(fcns)
        max_rows = pd.get_option('display.max_rows')  # get current setting
        pd.set_option('display.max_rows', len(s)) # set to len(s) so that all functions are printed.
        print(s.str.ljust(int(s.str.len().max())))
        pd.set_option('display.max_rows', max_rows)  # back to previous setting

    def get_available_chars(self):
        """Get the list of available characteristics.

        This function returns the aliases of the characteristics defined in `alias` column of ``mapping.xlsx``.

        Returns:
            A list of characteristics.
        """

        fcn_list = [fcn[2:] for fcn in dir(self) if callable(getattr(self, fcn)) and fcn[:2] == 'c_']
        char_list = [self.reverse_map[fcn] for fcn in fcn_list if fcn in self.reverse_map]
        return char_list

    def create_chars(self, char_list=None):
        """Generate firm characteristics.

        Created firm characteristics are added to ``data`` using their function names as the column names.

        Args:
            char_list: List of characteristics (aliases) to create. If None, all available characteristics are created.
        """

        char_list = char_list or self.get_available_chars()

        for i, char in enumerate(char_list):
            try:
                fcn = self.char_map[char]
                self.data[fcn] = eval(f'self.c_{fcn}()')
                elapsed_time(f"[{char}] created.")
                if i and (i % 30 == 0):
                    self.data = self.data.copy()
            except Exception as e:
                log(f'Error occured while creating {char}: {e}')
                raise e

    def prepare(self, char_fcns):
        """Prepare "ingredient" characteristics that are required to generate a characteristic.

        If a characteristic is defined as a function of other characteristics, this function will check whether they
        already exist and generate them if they do not exist.

        Args:
            char_fcns: list of function names (without 'c\_') of the ingredient characteristics.
        """

        for fcn in char_fcns:
            if fcn not in self.data:
                self.create_chars([self.reverse_map[fcn]])

    def add_row_count(self):
        """Add row count column ('rcount`) to `data`.

        The rcount has values 0, 1, ... for each ID starting from the first date.
        This can be use to remove the first n rows of each ID.

        NOTE:
            If the rows of `data` is changed, e.g., by filtering, this function should be called again as
            rcount may no longer be continuous.
        """

        self.data['rcount'] = self.data.groupby(self.id_col()).cumcount()

    def remove_rows(self, data, nrow=1):
        """Remove (set to nan) the first `nrow` rows of `data` for each ID.

        Args:
            data: Dataframe, Series, or ndarray. Its length should be the same as the length of `self.data`.
            nrow: Number of rows to remove (set to nan).

        Returns:
             `data` with rows removed.
        """

        m = int(self.freq / self.base_freq)
        data[self.data.rcount.values < nrow * m] = np.nan
        return data

    def populate(self, freq, method='ffill', limit=None):
        """Populate data.

        See ``datatools.populate()`` for details.
        """

        self.data = populate(self.data, freq, method, limit)
        self.freq = freq


    def copy(self, columns=None, deep=False):
        """Make a copy of this object and return it.

        Args:
            columns: Columns of `data` to copy.
            deep: If True, copy `data`, otherwise, only the address of `data` is copied.

        Returns:
            Copy of this object.
        """

        copy = Panel()
        copy.copy_from(self, columns, deep)
        return copy

    def copy_from(self, panel, columns=None, deep=False):
        """Copy `panel` to this object.

        Args:
            panel: `Panel` object to be copied.
            columns: Columns of `panel.data` to copy.
            deep: If True, copy `panel.data`, otherwise, only the address of `panel.data` is copied.
        """

        columns = columns or panel.data.columns
        self.data = panel.data[columns].copy(deep) if deep else panel.data[columns]
        self.freq = panel.freq
        self.base_freq = panel.base_freq
        self.char_map = panel.char_map
        self.reverse_map = panel.reverse_map

    def merge(self, right, on=None, how='inner', drop_duplicates='right'):
        """Merge `data` with `right`.

        Unlike `pd.merge()`, the index of `data` is always retained.

        Args:
            right: Series or Dataframe to merge with
            on: Columns to merge on. If None, merge on index.
            how: Merge method: 'inner', 'outer', 'left', or 'right'.
            drop_duplicates: how to handle duplicate columns. 'left': keep right, 'right': keep left,
                None: keep both. If None, '_x' ('_y') are added to the duplicate column names of left (right).
        """

        data = self.data
        index_names = data.index.names

        if on is None:
            is_index = True
            on = index_names
        elif isinstance(on, str):
            is_index = on in index_names
        else:
            is_index = False
            for i in on:
                is_index |= i in index_names

        if is_index:
            data.reset_index(inplace=True)

        dup_columns = data.columns.intersection(right.columns).difference(on)
        if drop_duplicates == 'right':
            self.data = data.merge(right[right.columns.difference(dup_columns)], on=on, how=how)
        elif drop_duplicates == 'left':
            self.data = data[data.columns.difference(dup_columns)].merge(right, on=on, how=how)
        else:
            self.data = data.merge(right, on=on, how=how)

        if is_index:
            self.data.set_index(index_names, inplace=True)

    def cumret(self, ret, period=1, lag=0):
        """Compute cumulative returns between t-`period` and t-`lag`.

        E.g., 12-month momentum can be obtained by setting `period` = 12 and `lag` = 1.
        A negative `period` will generate future returns: eg, `period` = -1 and `lag` = 0 for one-period ahead
        return; `period` = -3 and `lag` = -1 for two-period ahead
        return starting from t+1.

        NOTE:
            Returns `ret` should be in `base_freq`. If `base_freq` = ANNUAL, the returns in `ret` should be annual
            returns regardless of the value of `freq`.

        Args:
            ret: Series of returns in `base_freq`.
            period: Target horizon. (+) for past returns and (-) for future returns.
            lag: Period to calculate returns from.

        Returns:
            Series of cumulative returns.
        """

        if period > 0:
            if period == 1 and lag == 0:
                return ret.copy()
            else:
                return np.exp(self.rolling(np.log(ret + 1), period - lag, 'sum', lag=lag)) - 1
        else:
            if period == -1 and lag == 0:
                return self.shift(ret, -1)
            else:
                return np.exp(self.rolling(np.log(ret + 1), -(period - lag), 'sum', lag=period)) - 1

    def futret(self, ret, period=1):
        """Get `period`-period ahead return.
        """

        return self.cumret(ret, -period)

    def sum(self, *args):
        """Sum of `args`.

        This is similar to ``sum()`` of SAS: nan's of `args` are replaced by 0.

        Args:
            args: List of Series.

        Returns:
            (Series) Sum of `args`
        """

        y = np.zeros_like(args[0])
        isnan = True
        for x in args:
            y += x.fillna(0)
            isnan &= x.isna()
        y[isnan] = np.nan

        return y

    def shift(self, data, n=1):
        """Shift `data` accounting for frequency.

        This is similar to ``data.shift()`` but the actual shift period is determined by `freq` and `base_freq`.
        E.g., if `freq` = MONTHLY, `base_freq` = ANNUAL, `n` = 1 means 1-year shift, and `data`
        is shifted by 12 (12 months). That is, if `base_freq` = ANNUAL (QUARTERLY), ``shift(n)`` will always shift data
        by `n` years (quarters) regardless of the data frequency, `freq`.

        Args:
            data: Series or Dataframe.
            n: Shift size in terms of base frequency.

        Returns:
            Shifted data.
        """

        m = int(self.freq / self.base_freq)
        return self.remove_rows(data.shift(n * m), n)

    def pct_change(self, data, n=1, allow_negative_denom=False):
        """Get the percentage change of `data` accounting for frequency.

        This is similar to ``data.pct_change()`` but the actual period between two data points is determined by `freq` and `base_freq`.
        See ``Panel.shift()`` for details.

        Args:
            data: Series or Dataframe.
            n: Shift size in terms of base frequency.
            allow_negative_denom: If False, set the output to nan when the denominator is not positive.

        Returns:
            Series or Dataframe of percentage changes.
        """

        denom = self.shift(data, n)
        if not allow_negative_denom:
            denom[denom <= ZERO] = np.nan
        else:
            denom[denom == 0] = np.nan

        return data / denom - 1

    def diff(self, data, n=1):
        """Get the difference of the elements of `data` accounting for frequency.

        This is similar to ``data.diff()`` but the actual period between two data points is determined by `freq` and `base_freq`.
        See ``Panel.shift()`` for details.

        Args:
            data: Series or Dataframe.
            n: Shift size in terms of base frequency.

        Returns:
            Differenced data.
        """

        m = int(self.freq / self.base_freq)
        return self.remove_rows(data.diff(n * m), n)

    def _rolling(self, data, n, function, min_n=None, lag=0):
        if lag:
            roll = data.shift(lag).rolling(n, min_periods=min_n)
        else:
            roll = data.rolling(n, min_periods=min_n)

        if function == 'mean':
            return roll.mean()
        elif function == 'std':
            return roll.std()
        elif function == 'var':
            return roll.var()
        elif function == 'sum':
            return roll.sum()
        elif function == 'min':
            return roll.min()
        elif function == 'max':
            return roll.max()

    def rolling(self, data, n, function, min_n=None, lag=0):
        """Apply `function` to  a rolling window of `data`.

        This is similar to ``data.rolling()`` but the actual rolling window is determined by `freq` and `base_freq`.
        See ``Panel.shift()`` for details.

        Args:
            data: Series or Dataframe.
            n: Window size.
            function: Function to apply: 'mean', 'std', 'var', 'sum', 'min', or 'max'.
            min_n: Minimum number of observations in the window. If observations < `min_n`, result is nan.
                If `min_n` = None, `min_n` = `n`.
            lag: Lag size before start rolling.

        Returns:
            (Series or Dataframe) Rolling calculation result.

        Examples:
            Suppose `funda` is a `Panel` object with `freq` = MONTHLY and `base_freq` = ANNUAL (annual data populated
            monthly), and `funda.data` has a return column, 'ret'. The past three-year average return starting
            from one year ago can be calculated (with a condition of at least two non-missing values within the sample
            window) and saved as 'avg_ret3y' as follows:

            >>> funda.data['avg_ret3y'] = funda.rolling(funda.data['ret'], 3, 'mean', 2, 1)
        """

        m = int(self.freq / self.base_freq)
        if m == 1:
            return self.remove_rows(self._rolling(data, n, function, min_n, lag), n-1+lag)
        else:
            e = len(data)
            res = []
            for i in range(m):
                x = data.iloc[np.arange(i, e, m, dtype=int)]
                res.append(self._rolling(x, n, function, min_n, lag))

            return self.remove_rows(pd.concat(res).sort_index(level=[1, 0]), n-1+lag)

    def rolling_regression(self, y, X, n, min_n=None, add_constant=True, output=None):
        """Rolling OLS accounting for frequency.

        This function internally uses ``statsmodels.RollingOLS()``, which is slow when `output` is not None.
        Use ``Panel.rolling_beta()`` if possible.

        NOTE:
            When this function is used, the first `n-1` rows of each ID should be removed. This means that when
            `min_n` < `n`, some results can be lost. E.g., if `n` = 5 and `min_n` = 3, the rows with `data.rcount` < 4
            should be removed, and the results for `data.rcount` = 3, 4 will be lost.

        Args:
            y: Series of y with index = date/id.
            X: Series or Dataframe of X with index = date/id.
            n: Window size in terms of base frequency.
            min_n: Minimum number of observations in the window. If observations < `min_n`, result is nan.
                If `min_n` = None, `min_n` = `n`.
            add_constant: If True, add constant to `X`.
            output: List of outputs supported by ``statsmodels.RollingOLS()``. If None, estimate coefficients only.

        Returns:
            DataFrame of the coefficients if `output` = None, otherwise, DataFrames of the outputs defined by `output`.
        """

        m = int(self.freq / self.base_freq)

        params_only = output is None

        if add_constant:
            X = sm.add_constant(X)

        if m == 1:
            model = RollingOLS(y, X, window=n, min_nobs=min_n).fit(params_only=params_only)
            if output is None:
                return model.params
            else:
                return tuple([model.__getattribute__(arg) for arg in output])
        else:
            e = len(y)
            models = []
            params = []
            for i in range(m):
                y_ = y.iloc[np.arange(i, e, m, dtype=int)]
                X_ = X.iloc[np.arange(i, e, m, dtype=int)]
                model = RollingOLS(y_, X_, window=n, min_nobs=min_n).fit(params_only=params_only)
                models.append(model)
                params.append(model.params)

            if output is None:
                return self.remove_rows(pd.concat(params).sort_index(level=[1, 0]), n-1)
                # return pd.concat(params).reindex(y.index)
            else:
                retvals = []
                for arg in output:
                    retval = [model.__getattribute__(arg) for model in models]
                    retvals.append(self.remove_rows(pd.concat(retval).sort_index(level=[1, 0]), n-1))
                    # retvals.append(pd.concat(retval).reindex(y.index))

                return tuple(retvals)

    def rolling_beta(self, data, n, min_n=None):
        """Rolling OLS accounting for frequency.

        This is similar to ``Panel.rolling_regression()``: this function is faster but the output is limited to
        coefficients, R2, and idiosyncratic volatility.

        Args:
            data: DataFrame with index = date/id, sorted on id/date. The first column must be y and the rest X
                (not including constant).
            n: Window size in terms of base frequency.
            min_n: Minimum number of observations in the window. If observations < `min_n`, result is nan.
                If `min_n` = None, `min_n` = `n`.

        Returns:
            Coefficients, R2, idiosyncratic volatility. These are NxK, Nx1, and Nx1 ndarrays, respectively,
            where N = `len(data)` and K = number of X + 1.
        """

        m = int(self.freq / self.base_freq)

        if m == 1:
            return rolling_beta(data, n, min_n)
        else:
            e = data.shape[0]
            idx = []
            beta = []
            r2 = []
            idio = []
            for i in range(m):
                idx_ = np.arange(i, e, m, dtype=int)
                data_ = data.iloc[idx_]
                beta_, r2_, idio_ = rolling_beta(data_, n, min_n)
                idx.append(idx_)
                beta.append(beta_)
                r2.append(r2_)
                idio.append(idio_)

            idx = np.argsort(np.concatenate(idx))
            beta = np.concatenate(beta)[idx]
            r2 = np.concatenate(r2)[idx]
            idio = np.concatenate(idio)[idx]

            return beta, r2, idio





"""This module defines various toolkits for data-handling.

Rolling Application of Function
    The functions below apply a function to each group of a grouped data. For the differences,
    see the documentation of each function.

    * ``groupby_apply()``
    * ``groupby_apply_np()``
    * ``rolling_apply()``
    * ``rolling_apply_np()``

Grouping/Filtering/Trimming/Winsorization
    * ``classify()``: Classify data cross-sectionally.
    * ``filter_n()``: Filter data on a column.
    * ``filter()``: Filter data on a column with cut points based on another column.
    * ``trim()``: Trim data.
    * ``winsorize()``: Winsorize data.

Data Inspection/Comparison
    * ``inspect_data()``: Inspect a dataset.
    * ``compare_data()``: Compare two data sets.

Auxiliary Functions
    * ``populate()``: Populate data.
    * ``to_month_end()``: Shift dates to the last dates of the same month.
    * ``add_months()``: Add months to dates.
"""


import numpy as np
import pandas as pd
from pandas.tseries.offsets import MonthEnd

from pyanomaly.globals import *


################################################################################
#
# ROLLING APPLICATION OF FUNCTION
#
################################################################################
def _parse_retval(retval):
    """Parse returns from groupby_apply or groupby_apply_np.

    Args:
        retval: List of one of the followings:
            i) (tuple of) float
            ii) (tuple of) ndarray
            iii) (tuple of) dataframe/series.

    Returns:
        Concatenated value of `retval`. The number of returns is equal to the length of `retval[0]`.
    """

    if isinstance(retval[0], tuple):  # multiple returns
        n_ret = len(retval[0])
        retvals = [[] for _ in range(n_ret)]
        for retval_ in retval:
            for i in range(n_ret):
                retvals[i].append(retval_[i])

        if isinstance(retvals[0][0], float):
            for i in range(n_ret):
                retvals[i] = np.array(retvals[i]).reshape((-1,1))
        elif isinstance(retvals[0][0], np.ndarray):
            for i in range(n_ret):
                retvals[i] = np.concatenate(retvals[i])
        else:
            for i in range(n_ret):
                retvals[i] = pd.concat(retvals[i])

        return tuple(retvals)
    elif isinstance(retval[0], float):
        return np.array(retval).reshape((-1, 1))
    elif isinstance(retval[0], np.ndarray):
        return np.concatenate(retval)
    else:
        return pd.concat(retval)

def groupby_apply(gb, fcn, *args):
    """Apply `fcn` to groups in `gb`.

    Args:
        gb: pd.GroupBy object.
        fcn: Function to apply to groups. Function arguments should be `(gbdata, \*args)`,
            where `gbdata` is an element of `gb`.

    Returns:
        Concatenated outputs of `fcn`.
    """

    retval = [fcn(v, *args) for k, v in gb]
    return _parse_retval(retval)


def groupby_apply_np(gb, fcn, *args):
    """Apply `fcn` to the values of the groups in `gb`.

    This is the same as ``groupby_apply()`` except that the first argument of `fcn` is the values of the grouped data
    (ndarray). Use this function if `fcn` cannot receive DataFrame as an argument, e.g., when `fcn` is wrapped by
    ``numpy.jit``.

    NOTE:
        The groupby operation must not change the order of the data as the input to `fcn`, `data.values`, does not
        have an index. E.g., if `data` is sorted on `id`/`date`, grouping `data` by `id` is fine but not by `date`.

    Args:
        gb: pd.GroupBy object.
        fcn: Function to apply to groups. Function arguments should be `(gbdata_values, \*args)`,
            where `gbdata_values` is the values of an element of `gb`.

    Returns:
        Concatenated outputs of `fcn`.
    """

    retval = [fcn(v.values, *args) for k, v in gb]
    return _parse_retval(retval)


def rolling_apply(data, by=None, fcn=None, *args):
    """Apply `fcn` to each group of data grouped by `by`.

    This is similar to ``groupby_apply()``: ``groupby_apply()`` receives pd.GroupBy object as input whereas this
    function performs ``pd.groupby()`` inside.

    Args:
        data: DataFrame to be grouped.
        by: Column by which `data` is grouped. If None, `data` is grouped by the last index.
        fcn: Function to apply to groups.  Function arguments should be `(gbdata, \*args)`,
            where `gbdata` is an element of the grouped data.

    Returns:
        Concatenated outputs of `fcn`.
    """

    by = by or data.index.names[-1]
    return groupby_apply(data.groupby(by), fcn, *args)


def rolling_apply_np(data, by=None, fcn=None, *args):
    """Apply `fcn` to each group of data grouped by `by`.

    This is the same as ``rolling_apply()`` except that the first argument of `fcn` is the values of the grouped data
    (ndarray). Use this function if `fcn` cannot receive DataFrame as an argument, e.g., when `fcn` is wrapped by
    ``numpy.jit``.

    NOTE:
        The groupby operation must not change the order of the data as the input to `fcn`, `data.values`, does not
        have an index. E.g., if `data` is sorted on `id`/`date`, grouping `data` by `id` is fine but not by `date`.

    Args:
        data: DataFrame to be grouped.
        by: Column by which `data` is grouped. If None, `data` is grouped by the last index, i.e., groupby(level=-1).
        fcn: Function to apply to groups. Function arguments should be `(gbdata_values, \*args)`,
            where `gbdata_values` is the values of an element of the grouped data.


    Returns:
        Concatenated outputs of `fcn`.
    """

    by = by or data.index.names[-1]
    return groupby_apply_np(data.groupby(by), fcn, *args)


# @staticmethod
# # @multiprocessing
# def rolling_apply1(cd, fcn, n_retval):
#     """Apply fcn in every month. This function runs a loop over permno and can be used when the calculation inside
#     fcn requires data over several months.
#
#     Args:
#         cd: crspd.data
#         data_columns: columns of cd that are used as input to fcn.
#         fcn(data): function to apply to a particular permno.
#             data: crspd.data[columns] for a given permno, where columns indicate the data used in fcn.
#             returns: num_months x num_returns ndarray.
#         n_retval: number of returns from fcn.
#
#     Returns:
#         (permno/mon) x num_returns ndarray.
#     """
#     ngroups = cd.groupby(['permno', 'mon']).ngroups
#     gsize = cd.groupby(['permno']).size().values
#     data = cd.values
#     mons = cd['mon'].values
#     # isnan = np.isnan(data).any(axis=1)
#
#     @njit
#     def rolling_apply1_(data, gsize, ngroups, mons):
#         retval = np.full((ngroups, n_retval), np.nan)
#
#         j, idx0 = 0, 0
#         for i in range(gsize.shape[0]):
#             data_ = data[idx0:idx0 + gsize[i]]
#             n_mon = np.unique(mons[idx0:idx0 + gsize[i]]).shape[0]
#
#             retval_ = fcn(data_)
#             if n_retval == 1:
#                 retval[j:j + n_mon, 0] = retval_
#             else:
#                 for k in range(n_retval):
#                     retval[j:j + n_mon, k] = retval_[k]
#
#             idx0 += gsize[i]
#             j += n_mon
#
#         return retval
#
#     return rolling_apply1_(data, gsize, ngroups, mons)



################################################################################
#
# DATA GROUPING/FILTERING/TRIM/WINSORIZE
#
################################################################################

def _combine_array(array, by_array):
    if by_array is not None:
        return pd.DataFrame(np.array([array.values, by_array.values]).T)
    else:
        return array.reset_index(drop=True)


def _split_apply(array, fcn, *args):
    if array.ndim == 1:
        return fcn(array, *args)
    else:
        return fcn(array.iloc[:, 0], *args, array.iloc[:, 1])


def classify_array(array, split, ascending=True, by_array=None):
    """Classify array into classes.

    Class labels are set to 0, ..., no. quantiles-1, where 0 corresponds to the lowest (highest) group
    if `ascending` = True (False).
    If array contains nan, their classes are set to nan.

    Args:
        array: Nx1 ndarray or Series to be grouped.
        split: Number of classes (for equal-size quantiles) or list of quantiles. (0.3, 0.7) is equivalent to (0.3, 0.7, 1.0).
        ascending (bool): Sorting order.
        by_array: Array based on which cut points are determined. If None, by_array = array. Eg, array can be set
            to ME and by_array to NYSE-ME to group firms on size with NYSE-size cut points.

    Returns:
        Nx1 ndarray or Series of classes. The output type corresponds to the input type.

    Examples:
        .. code-block::

            split = 10  # Classify array into 10 equally-spaced groups (0, ..., 9).
            split = [0.3, 0.7, 1.0]  # Classify array into three groups (0, 1, 2) that correspond to 0.3, 03-0.7, 0.7-1.0
                                     # quantiles.
            ascending = False  # Label 0 represents the largest group.
    """

    if isinstance(split, (int, np.integer)):  # number of classes
        quantiles = np.linspace(0, 1, split + 1)[1:]
    else:  # quantiles
        quantiles = split + ([1.0] if split[-1] < 1 else [])

    classes = np.full(array.shape, np.nan)

    if by_array is None:
        by_array = array
    by_array = by_array[~np.isnan(by_array)]

    if len(by_array) == 0:
        return pd.Series(classes, index=array.index) if type(array) == pd.Series else classes

    values = np.quantile(by_array, quantiles[:-1])
    classes[array > values[-1]] = 0
    for i, v in enumerate(values[::-1]):
        classes[array <= v] = i + 1

    if ascending:
        classes = len(quantiles) - 1 - classes

    return pd.Series(classes, index=array.index) if type(array) == pd.Series else classes


def classify(array, split, ascending=True, by_array=None):
    """ Classify `array` cross-sectionally.

    Class labels are set to 0, ..., no. quantiles-1, where 0 corresponds to the lowest (highest) group
    if `ascending` = True (False). If `array` contains nan, their classes are set to nan.

    Args:
        array: Data to classify. Series with date/id index.
        split: Number of classes or list of quantiles (see `classify_array()`).
        ascending (bool): Sorting order.
        by_array: Data based on which quantile breakpoints are determined. If None, `by_array` = `array`.

    Returns:
        Series of class labels with index = `data.index`.

    Examples:
        Suppose `data` is a panel data that has columns 'me' (market equity) and 'me_nyse' (same as 'me' but
        has a value only when the stock is listed on NYSE).
        The following code groups stocks into terciles based on size with NYSE-size breakpoints and store the output
        in `data['me_class']`.
    ::

        split = [0.2, 0.7, 1.0]  # bottom 20%, 20-70%, top 30%
        data['me_class'] = classify(data['me'], split, by=data['me_nyse'])
    """

    array_ = _combine_array(array, by_array)
    classes = groupby_apply(array_.groupby(array.index.get_level_values(0)), _split_apply, classify_array, split, ascending)
    classes.sort_index(inplace=True)  # Restore the order of the rows.
    classes.index = array.index
    return classes


def _filter(array, n=None, q=None, ascending=True):
    v = array.values
    nancnt = int(np.isnan(v).sum())
    n = int(q * (len(v)-nancnt)) if q else n  # number of rows to keep

    idx = np.full(len(v), False)
    if n and ascending:
        idx[np.argpartition(v, min(len(v) - 1, n))[:n]] = True
    elif n:
        idx[np.argpartition(v, max(len(v) - n - nancnt, 0))[-n - nancnt:]] = True
        idx[np.isnan(v)] = False
    return pd.Series(idx, index=array.index)


def filter_n(data, on, n=None, q=None, ascending=True):
    """Filter `data` on `on`.

    If `n` is given, choose `n` smallest (if `ascending` = True) or largest (if `ascending` = False) within
    each date. If `q` is given, choose `q` quantile (from the bottom if `ascending` = True) within each date.

    Args:
        data: DataFrame or Series. The first index should be date.
        on: Column to filter on.
        n: Number of rows to keep.
        q: Quantile to keep.
        ascending (bool): If True, keep the smallest.

    Returns:
        Filtered data.

    Examples:
        Suppose `data` is a panel data (index = date/id) and has a column 'me' (market equity).
        The smallest 20% can be filtered out (keep the largest 80%) as follows:

        >>> data = filter(data, 'me', 0.8, ascending=False)
    """
    if (n is None) == (q is None):
        raise ValueError(f'Only one of n or q should be given.')

    array = data[on].reset_index(drop=True)
    idx = groupby_apply(array.groupby(data.index.get_level_values(0)), _filter, n, q, ascending)
    idx.sort_index(inplace=True)  # Restore the order of the rows.
    return data[idx.values]


def filter(data, on, limits, by=None):
    """Filter `data` on `on` with cut points determined by `by`.

    Filter out data if `data[on]` is below `limits[0]` quantile or above (1 - `limits[1]`) quantile.
    This function is similar to ``filter_n()`` but the cut points can be determined by another column values instead
    of the values of `on`.

    Args:
        data: DataFrame or Series. The first index should be date.
        on: Column to filter on.
        limits: Pair of quantiles, eg, (0.1, 0.1), (0.1, None).
        by: Column based on which cut points are determined. If None, `by` = `col`.

    Returns:
        Filtered data.

    Examples:
        Suppose `data` is a panel data (index = date/id) and has columns 'me' (market equity) and 'nyse_me'
        (market equity of NYSE stocks). The stocks whose sizes are smaller than the 0.2 NYSE-size quantile can be
        removed as follows:

        >>> data = filter(data, 'me', (0.2, None), by='nyse_me')
    """

    by = by or on
    array = data[on]
    by_array = data[by]

    array_ = _combine_array(array, by_array)

    idx = groupby_apply(array_.groupby(array.index.get_level_values(0)), _split_apply, _trim1d, limits)
    idx.sort_index(inplace=True)  # Restore the order of the rows.
    return data[idx.values]


def _trim1d(array, limits, by_array=None):
    if by_array is None:
        by_array = array
    by_array = by_array[~np.isnan(by_array)]

    lq = np.quantile(by_array, limits[0]) if limits[0] else None
    uq = np.quantile(by_array, 1-limits[1]) if limits[1] else None

    return (array > lq if lq else True) & (array < uq if uq else True)


def trim(array, limits, by_array=None):
    """Trim `array` within each date that are outside of the quantile values defined by `limits`.

    E.g., if `limits` = (0.1, 0.1), trim data below 0.1 quantile and above 0.9 quantile. The elements of `limits` can be
    set to None for one-sided trim, e.g., `limits` = (0.1, None).

    Args:
        array (Series): Data to trim. The first index should be date.
        limits: Pair of quantiles, eg, (0.1, 0.1), (0.1, None).
        by_array (Series): Data based on which cut points are determined. If None, `by_array` = `array`.

    return:
        Trimmed data.
    """

    array_ = _combine_array(array, by_array)
    idx = groupby_apply(array_.groupby(array.index.get_level_values(0)), _split_apply, _trim1d, limits)
    idx.sort_index(inplace=True)  # Restore the order of the rows.
    return array[idx.values]


def _trim2d(data, limits):
    lq = data.quantile(limits[0]) if limits[0] else None
    uq = data.quantile(1-limits[1]) if limits[1] else None

    for col in data:  # Set out-of-bound values to nan.
        data.loc[(True if lq is None else data[col] < lq[col]) |
                 (True if uq is None else data[col] > uq[col]), col] = np.nan

    return data

def trim_data(data, limits):
    """Trim each column of `data` within each date that are outside of the quantile values defined by `limits`.

        The trimmed values are set to nan and the output shape is equal to the input shape.

    Args:
        data (DataFrame): Data to trim. The first index should be date.
        limits: Pair of quantiles, eg, (0.1, 0.1), (0.1, None).

    return:
        Trimmed data.
    """

    index = data.index
    data = data.reset_index(drop=True)
    data = groupby_apply(data.groupby(index.get_level_values(0)), _trim2d, limits)
    data.sort_index(inplace=True)  # Restore the order of the rows.
    data.index = index
    return data


def _winsorize(data, limits, by_array=None):
    if by_array is None:
        by_array = data
    lq = by_array.quantile(limits[0]) if limits[0] else None
    uq = by_array.quantile(1-limits[1]) if limits[1] else None

    if data.ndim == 1:
        return data.clip(lower=lq, upper=uq)
    else:
        return data.clip(lower=lq, upper=uq, axis=1)

def winsorize(array, limits, by_array=None):
    """Winsorize `array` within each date that are outside of the quantile values defined by `limits`.

    Args:
        array (Series): Data to winsorize. The first index should be date.
        limits: Pair of quantiles, eg, (0.1, 0.1), (0.1, None).
        by_array (Series): Data based on which cut points are determined. If None, `by_array` = `array`.

    Returns:
        Winsorized data.
    """

    array_ = _combine_array(array, by_array)
    new_array = groupby_apply(array_.groupby(array.index.get_level_values(0)), _split_apply, _winsorize, limits)
    new_array.sort_index(inplace=True)  # Restore the order of the rows.
    new_array.index = array.index
    return new_array

def winsorize_data(data, limits):
    """Winsorize each column of `data` within each date that are outside of the quantile values defined by `limits`.

    Args:
        data (DataFrame): Data to winsorize.. The first index should be date.
        limits: Pair of quantiles, eg, (0.1, 0.1), (0.1, None).

    Returns:
        Winsorized data.
    """

    index = data.index
    data = data.reset_index(drop=True)
    data = groupby_apply(data.groupby(index.get_level_values(0)), _winsorize, limits)
    data.sort_index(inplace=True)  # Restore the order of the rows.
    data.index = index
    return data


################################################################################
#
# DATA INSPECTION/COMPARISON
#
################################################################################

def inspect_data(data, option=['summary'], date_col=None, id_col=None):
    """Inspect data.

    This function inspects a panel data, `data`, and print the results.

    * 'summary': `data` shape, number of unique dates, and number of unique ids.
    * 'id_count`: Number of ids per date.
    * 'nans': Number of nans and infs per column.
    * 'stats': Descriptive statistics. Same as ``data.describe()``.

    Args:
        data: Dataframe. It should have date and id columns or index = date/id.
        option: List of items to display. Available options:
                'summary', 'id_count', 'nans', 'stats'.
        date_col: Date column. If None, `date.index[0]` is assumed to be date.
        id_col: ID column. If None, `date.index[1]` is assumed to be id.
    """

    max_rows = pd.get_option('display.max_rows')  # get current setting
    pd.set_option('display.max_rows', data.shape[1])  # set to num. of columns.

    if date_col:
        dates = data[date_col]
    else:
        date_col = data.index.names[0]
        dates = data.index.get_level_values(0)

    if id_col:
        ids = data[id_col]
    else:
        ids = data.index.get_level_values(1)

    if 'summary' in option:
        log(f'Shape: {data.shape}')
        log(f'No. unique dates: {len(dates.unique())}, Date range: {dates.min()} - {dates.max()}')
        log(f'No. unique ids: {len(ids.unique())}')

    if 'id_count' in option:
        log(f'No. ids per date:\n {data.groupby(date_col).size()}')

    if 'nans' in option:
        log('\n')
        log('Nans, Infs')
        nans = data.isna().sum()
        nans = pd.concat([nans, nans / data.shape[0]], axis=1, keys=['no. nans', 'percentage'])

        infs = ((data == np.inf) | (data == -np.inf)).sum()
        infs = pd.concat([infs, infs / data.shape[0]], axis=1, keys=['no. infs', 'percentage'])

        log(pd.concat([nans, infs], axis=1), header=False)

    if 'stats' in option:
        log('\n')
        log('Descriptive statistics')
        log(data.describe([0.001, 0.01, .25, .5, .75, 0.99, 0.999]).T, header=False)

    pd.set_option('display.max_rows', max_rows)  # back to previous setting


def compare_data(data1, data2=None, on=None, how='inner', tolerance=0.01, suffixes=('_x', '_y'), returns=True):
    """Compare `data1` with `data2`.

    This function compares the common columns of `data1` and `data2`.
    This is similar to ``data1.compare(data2)``, but `data1` and `data2` are not required to have the same index and
    columns. Also, a tolerance can be set to determine whether two values are the same.

    Args:
        data1: Dataframe for comparison.
        data2: Dataframe for comparison. If None, `data1` is assumed to be a merged dataset of `data1` and `data2`.
            If `data1` is a merged dataset, `on` and `how` have no effect.
        on: A column or a list of columns to merge data sets on. If None, data sets will be merged on index.
        how: How to merge: 'inner', 'outer', 'left', or 'right'
        tolerance: Tolerance level to determine equality. Two values, `val1` and `val2`, are considered to be the same
            if `abs((val1 - val2) / val2) < tolerance`.
        suffixes: suffixes to add to common columns or suffixes used in the merged dataset.
            `suffixes[0]`: suffix for `data1`, `suffixes[1]`: suffix for `data2`.
        returns: If True, return the merged data.
    """

    if data2 is None:  # data1 is already a merged dataset.
        data = data1
    else:
        on = on or data1.index.names  # if None, merge on index.
        data = pd.merge(data1, data2, on=on, how=how, suffixes=suffixes)
        data.sort_index(axis=1, inplace=True)

    log(f"{'column': <20}  {'matched': >7}  {'corr': >7}", header=False)
    for col in data.columns:
        if (col[-2:] == suffixes[0]):
            if data[col].dtype not in ('float', 'float32', 'int', 'int64'):
                continue
            col2 = col[:-2] + suffixes[1]
            if col2 not in data:
                continue
            pair = data[[col, col2]]
            pair = pair[~pair.isin([np.nan, np.inf, -np.inf]).any(1)]  # drop nan or inf
            corr = pair[col].corr(pair[col2])  # np.corrcoef(pair.T)[0, 1]
            pair['diff'] = (pair[col] - pair[col2]).abs()
            matchidx = ((pair[col2].abs() < 1e-5) & (pair['diff'] < 1e-5)) | (pair['diff'] / pair[col2].abs() < tolerance)
            match = matchidx.sum() / len(pair)
            log(f'{col[:-2]: <20} {match: 7.5f} {corr: 7.5f}', header=False)

    if returns:
        return data


################################################################################
#
# AUX FUNCTIONS
#
################################################################################

from concurrent.futures import ProcessPoolExecutor, as_completed
from multiprocessing import cpu_count

def _fcn(k, g, freq, method, limit):
    if method == 'ffill':
        return k, g.resample(freq_map[freq]).pad(limit)
    else:  # fill with None
        return k, g.resample(freq_map[freq]).asfreq()


def populate(data, freq, method='ffill', limit=None):
    """Populate data.

    References:
        https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.resample.html

    Args:
        data: Dataframe with index = date/id.
        freq: Frequency to populate: QUARTERLY, MONTHLY, or DAILY.
        method: Filling method for newly added rows. 'ffill': forward fill, None: None.
        limit: Maximum number of rows to forward-fill.

    Returns:
        Populated data.
    """
    id_col = data.index.names[-1]

    gb = data.reset_index(level=id_col).groupby(id_col)
    retval = []
    futures = []
    with ProcessPoolExecutor() as executor:
        for k, g in gb:
            futures.append(executor.submit(_fcn, k, g, freq, method, limit))

        for f in as_completed(futures):
            k, retval_ = f.result()
            retval_[id_col] = k
            retval.append(retval_)

    return pd.concat(retval).set_index(id_col, append=True).sort_index(level=[1,0])


# def populate2(data, freq, method='ffill', limit=None):
#     """Populate data.
#
#     References:
#         https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.resample.html
#
#     Args:
#         data: Dataframe with index = date/id
#         freq: Frequency to populate: QUARTERLY, MONTHLY, or DAILY.
#         method: Filling method for newly added rows. 'ffill': forward fill, None: None.
#         limit: Maximum number of rows to forward-fill.
#
#     Returns:
#         Populated data.
#     """
#
#     id_col = data.index.names[-1]
#
#     gb = data.reset_index(level=id_col).groupby(id_col)
#     retval = []
#     for k, g in gb:
#         if method == 'ffill':
#             retval.append(g.resample(freq_map[freq]).pad(limit))
#         else:  # fill with None
#             retval.append(g.resample(freq_map[freq]).asfreq())
#         retval[-1][id_col] = k
#
#     return pd.concat(retval).set_index(id_col, append=True)
#
#     # The code below is simpler but three times slower.
#     # if method == 'ffill':
#     #     pop_data = data.reset_index(level=id_col).groupby(id_col).resample(freq_map[freq]).pad(limit).swaplevel()
#     # else:  # fill with None
#     #     pop_data = data.reset_index(level=id_col).groupby(id_col).resample(freq_map[freq]).asfreq().swaplevel()
#     # pop_data.drop(columns=id_col, inplace=True)
#     #
#     # return pop_data


def to_month_end(date):
    """Shift dates to the last dates of the same month.

    Args:
        date: datetime Series.

    Returns:
        datetime Series shifted to month end.
    """

    return date + MonthEnd(0)


def add_months(date, months, to_month_end=True):
    """Add months to dates.

    Args:
        date: datetime Series.
        months: Months to add. Can be negative.
        to_month_end: If True, returned dates are end-of-month dates.

    Returns:
        datetime Series of (`date` + `months`). Dates are end-of-month dates if `to_month_end` = True.
    """

    if to_month_end:
        return date + MonthEnd(months)
    else:
        return date + pd.DateOffset(months=months)




